/** 
 * @file Initialise et configure le serveur Express sur lequel tourne le back.
 * 
 * La configuration inclut tout le _middleware_ définissant les API et les services
 * nécessaire utilisés, comme `express-session`, GraphiQL, GraphQL Voyager.
 * @author manifold
*/
import express from 'express';
import schema from './graphql/schema';
import { express as graphqlVoyager } from 'graphql-voyager/middleware';
import graphqlHTTP from 'express-graphql'; // new name of 'graphql-server-express'. cf npmjs.com
import flash from 'connect-flash';
import { ensureLoggedIn } from 'connect-ensure-login';
import passport from 'passport';
import LdapStrategy from 'passport-ldapauth';
import fs from 'fs';
import session from 'express-session';
import bodyParser from 'body-parser';
import favicon from 'serve-favicon';
import morgan from 'morgan';
import path from 'path';
import cors from 'cors';

const server = express();

// Parse incoming HTTP request bodies, available under the req.body property
// cf www.npmjs.com/package/body-parser
server.use(bodyParser.json()); //parses bodies of media type "application/json"
server.use(bodyParser.urlencoded({ //parses bodies of media type "application/x-www-form-urlencoded"
    extended: true //use qs library (quoi que ca veuille dire o.O)
}));

/**
 * @description Configuration de l'authentification
 * @author guillaume.wang
 *
 * on a besoin d'authentification pour 2 trucs :
 * - l'acces a l'interface admin (admin_view) du back, definie dans admin_router.js
 * - le contexte graphQL
 * 
 * serializeUser et deserializeUser: passport s'attend a ce qu'on ait besoin d'avoir req.user disponible partout dans notre code
 * En gros l'idee de passport c'est: serializeUser permet d'obtenir une cle identifiant chaque user
 * et deserializeUser prend cette cle, fait une requete vers une BDD de users et met dans l'objet JS req.user toutes les infos issues de la BDD
 * Cette repartition permet de ne stocker dans la session (i.e. en memoire sur le serveur) que la cle des utilisateurs connectes et de ne "charger en memoire" toutes les infos de la BDD que lorsque necessaire
 * cf https://stackoverflow.com/questions/27637609/understanding-passport-serialize-deserialize#27637668
 * et http://toon.io/understanding-passportjs-authentication-flow/
 * 
 * Mais en fait dans notre cas c'est graphql qui communique avec la BDD, donc on s'en fiche! On peut se contenter de dire a serializeUser et deserializeUser de ne s'occuper que du champ uid)
 */
let configPath = path.resolve('./', 'ldap_config.json');
let config = JSON.parse(fs.readFileSync(configPath, 'utf8'));

passport.use(new LdapStrategy({
    server: {
        url: config.ldap.server,
        //bindDn: '.............',
        //bindCredentials: '..........',
        searchBase: config.ldap.searchBase,
        searchFilter: config.ldap.searchFilter,
        //searchAttributes: ['givenName', 'sn'],
        //tlsOptions: '..........',
    },

    //usernameField: 'username', // Field name where the username is found, defaults to username
    //passwordField: 'password', // Field name where the pas    sword is found, defaults to password
    
    // LdapStrategy has a default verify callback ! j'ai perdu plein de temps pour rien :'(
    // cf. https://github.com/vesse/passport-ldapauth/blob/master/lib/passport-ldapauth/strategy.js, line 195 (` var verify = function() { ... } `)
    /*
    function (user, done) {
        // "verify callback", called after each passport.authenticate(...),
        // unless missing credentials (in which case a 400 Error is returned)

        // "The purpose of a verify callback is to find the user that possesses a set of credentials" (from passport doc)
        // i.e. we query the database (in our case the LDAP) to get user's data
        console.log("Entering passport's verify callback");

        if (user){
            //if user exists
            console.log("Successfully authenticated " + user.uid);
        }
    }
    */
})
);

// Définit les paramètres de stockage des sessions.
server.use(session({
    secret: config.sessionSecret,
    resave: true,
    saveUninitialized: false
}));
server.use(passport.initialize());
server.use(passport.session());

//toujours bon a savoir pour faire des tests:
//The result of the serializeUser method is attached to the session as req.session.passport.user
passport.serializeUser(function (user, done) {
    done(null, user.uid);
});

//The first argument of deserializeUser corresponds to the key of the user object that was given to the done function in serializeUser
//The fetched object is attached to the request object as req.user (available in all subsequent middleware)
passport.deserializeUser(function (userUid, done) {
    done(null, { uid: userUid });
});
/* fin de Configuration de l'authentification */

// cache le fait que l'application tourne sous Express dans le header HTTP.
server.disable('x-powered-by');

// setting up view engine for pug
console.log("Running at",__dirname);
let viewpath = path.resolve(__dirname,'views');
server.set('views', viewpath);
server.set('view engine', 'pug');

// favicon: capital sigma symbol
server.use(favicon(path.resolve('./','assets','favicon.ico')));

// specifies path to static assets
server.use('/assets',express.static(path.resolve('./','assets')));

// Morgan is middleware for logging requests
server.use(morgan('dev'));

const defaultUser = require('./../ldap_connexion_config.json');

// Options de configuration pour le _middleware_ `cors`.
// CORS = Cross Origin Resource Sharing
const corsOptions = {
    origin: 'http://localhost:8888', // Configures the Access-Control-Allow-Origin CORS header. i.e. specifies that sigma-back wants to make resources accessible to this site (and this site only)
    credentials: true // Configures the Access-Control-Allow-Credentials CORS header. i.e. allows cookies to be included on cross-origin requests
};
server.use(cors(corsOptions));

/*
server.use('/graphql', bodyParser.json(), graphqlHTTP(async (req, res, params) => ({
    schema: schema,
    graphiql: true,
    context: {
        user: {
            uid: req.user ? req.user.uid : defaultUser.dn.split("=")[1].split(",")[0],
            password: "mythe"
        }
    }
})));
*/

server.use('/graphql', 
    bodyParser.json(), // parse incoming HTTP request (req) as a JSON
    graphqlHTTP(async (req, res, params) => {
        // vary the options *on a per-request basis*
        let uid;
        let password;

        if(req.isAuthenticated) {
            try {
                uid = req.user.uid;
                password = "mythe";
            } catch (err) {
                console.log("Error: req is authenticated, but pb when trying to extract uid from req.user. Probably user was either not serialized or not deserialized properly");
                console.log(err);
            }
        } else {
            // FOR DEVELOPMENT ONLY. for production, replace with a "publicUser" or "notLoggedInUser" or something.
            uid = defaultUser.dn.split("=")[1].split(",")[0];
            password = defaultUser.passwd;
        }
        // console.log("Accessing GraphQL as: ",uid);

        return {
            schema: schema,
            graphiql: true, // gives access to graphiql if request is detected to be from browser (je crois)
            context: { user: { uid: uid, password: password } } // accessible in every single resolver as the third argument
        };
    })
);

// GraphQL voyager affiche une représentation sous forme de graphe du schema GraphQL
server.use('/voyager',
    /*ensureLoggedIn('/login'),*/
    graphqlVoyager({ endpointUrl: '/graphql' })
);

// connect-flash is middleware for flashing messages
// used in sigma-back's admin interface (admin_view)
server.use(flash());

export default server;
